##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging

from typing import Optional

from azure.quantum.target.solvers import Solver
from azure.quantum.workspace import Workspace

logger = logging.getLogger(__name__)

class Tabu(Solver):
    target_names = (
        "microsoft.tabu.cpu",
        "microsoft.tabu-parameterfree.cpu"
    )
    def __init__(
        self,
        workspace: Workspace,
        name: str = "microsoft.tabu.cpu",
        sweeps: Optional[int] = None,
        tabu_tenure: Optional[int] = None,
        timeout: Optional[int] = None,
        seed: Optional[int] = None,
        restarts: Optional[int] = None,
        **kwargs
    ):
        """The constructor of an Tabu Search solver.

        Multi-core Tabu Search solver for binary optimization problems
        with k-local interactions on an all-to-all graph topology with double
        precision support for the coupler weights.

        This solver is CPU only.

        :param sweeps:
            specifies the number of sweeps.
        :param tabu_tenure:
            specifies the tabu tenure.
        :param restarts:
            specifies how many runs the solver will execute.
        :param timeout:
            specifies maximum number of seconds to run the core solver
            loop. initialization time does not respect this value, so the
            solver may run longer than the value specified.
        :param seed:
            specifies a random seed value.
        """
        param_free = sweeps is None and tabu_tenure is None and restarts is None
        if param_free:
            name = name.replace(".tabu.", ".tabu-parameterfree.")

        super().__init__(
            workspace=workspace,
            provider_id="Microsoft",
            name=name,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2",
            **kwargs
        )

        self.set_one_param("sweeps", sweeps)
        self.set_one_param("tabu_tenure", tabu_tenure)
        self.set_one_param("timeout", timeout)
        self.set_one_param("seed", seed)
        self.set_one_param("restarts", restarts)

    def supports_grouped_terms(self):
        return True
    
    def supports_protobuf(self):
        return True


##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import abc
from typing import Any, Dict


from azure.quantum.aio.job.job import Job
from azure.quantum.target import Target as SyncTarget
from azure.quantum.aio.workspace import Workspace


class Target(SyncTarget, abc.ABC):
    """Azure Quantum Target."""
    workspace: Workspace

    async def submit(
        self,
        input_data: Any,
        name: str = "azure-quantum-job",
        input_params: Dict[str, Any] = None,
        **kwargs
    ) -> Job:
        """Submit input data and return Job

        :param input_data: Input data
        :type input_data: Any
        :param name: Job name
        :type name: str
        :param input_params: Input parameters
        :type input_params: Dict[str, Any]
        :return: Azure Quantum job
        :rtype: Job
        """
        input_params = input_params or {}
        input_data_format = kwargs.pop("input_data_format", self.input_data_format)
        output_data_format = kwargs.pop("output_data_format", self.output_data_format)
        content_type = kwargs.pop("content_type", self.content_type)
        encoding = kwargs.pop("encoding", self.encoding)
        blob = self._encode_input_data(data=input_data)

        return await Job.from_input_data(
            workspace=self.workspace,
            name=name,
            target=self.name,
            input_data=blob,
            content_type=content_type,
            encoding=encoding,
            provider_id=self.provider_id,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            input_params=input_params,
            **kwargs
        )

    async def refresh(self):
        """Update the target availability and queue time"""
        targets = await self.workspace._get_target_status(self.name, self.provider_id)
        if len(targets) > 0:
            _, target_status = targets[0]
            self._current_availability = target_status.current_availability
            self._average_queue_time = target_status.average_queue_time
        else:
            raise ValueError(
                f"Cannot refresh the Target status: \
target '{self.name}' of provider '{self.provider_id}' not found."
            )



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from __future__ import annotations
import numpy as np
from typing import List, Dict, Union, Optional
from enum import Enum
from abc import ABC

__all__ = ["TermBase", "Term", "GroupType", "SlcTerm"]

try:
    import numpy.typing as npt

    WArray = Union[int, float, npt.ArrayLike]

    def _convert_if_numpy_type(param: WArray):
        # Attempt first a conversion to a supported
        # type if parameter is a numpy float/int.
        numpy_integer_types = [
            np.byte,
            np.ubyte,
            np.short,
            np.ushort,
            np.intc,
            np.uintc,
            np.int_,
            np.uint,
            np.longlong,
            np.ulonglong,
            np.int8,
            np.uint8,
            np.int16,
            np.uint16,
            np.int32,
            np.uint32,
            np.int64,
            np.uint64,
        ]

        numpy_float_types = [
            np.float16,
            np.float32,
            np.float64,
            np.float_,
            np.half,
            np.single,
            np.double,
        ]

        if hasattr(param, "__iter__"):
            # Handle scalar-like arrays, if specified.
            param = param[0]

        if (
            hasattr(param, "dtype")
            and param.dtype in numpy_integer_types + numpy_float_types
        ):
            return param.item()
        else:
            return param


except ImportError:
    npt = None
    WArray = Union[int, float]

    def _convert_if_numpy_type(param: WArray):
        return param


class TermBase(ABC):
    """
    Term base class; this class is not directly initialized
    """
    def __init__(
        self,
        c: Optional[WArray] = None,
    ):
        if c is not None:
            # Current intended specification of term.
            coeff = c
        else:
            raise RuntimeError("Cost should be provided for each term.")

        coeff = _convert_if_numpy_type(coeff)
        if type(coeff) != int and type(coeff) != float:
            raise RuntimeError(
                "c must be a float or int value, \
                    or a NumPy value that can be converted to those."
            )
        self.c = coeff

    def to_dict(self):
        return self.__dict__

    @classmethod
    def from_dict(cls, obj):
        return cls(c=obj.get("c"))

    def evaluate(self, *args, **kwargs) -> float:
        """Given a variable configuration, evaluate the value of the term.
        :param configuration:
            The dictionary of variable ids to their assigned value
        """
        return self.c

    def reduce_by_variable_state(self, *args, **kwargs) -> Optional[TermBase]:
        """Given some fixed variable states,
            transform the existing term into new term.
        Returns None if the new term is effectively 0
        :param fixed_variables:
            The dictionary of variable ids and their fixed state
        """
        return TermBase(c=self.c) if self.c != 0 else None

    def __repr__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

class Term(TermBase):
    """
    Class describing a single (monomial) term.
    """
    def __init__(
        self,
        indices: List[int] = None,
        w: Optional[WArray] = None,
        c: Optional[WArray] = None,
    ):
        if w is not None:
            # Legacy support if 'w' is used to specify
            # term instead of the expected 'c'.
            coeff = w
            parameter_name_used = "w"
        elif c is not None:
            coeff = c
            parameter_name_used = "c"
        else:
            raise RuntimeError("Cost should be provided for each term.")
        coeff = _convert_if_numpy_type(coeff)
        if type(coeff) != int and type(coeff) != float:
            raise RuntimeError(
                f"{parameter_name_used} must be a float or int value, \
                    or a NumPy value that can be converted to those."
            )
        TermBase.__init__(self, c=coeff)
        self.ids = indices
    
    @classmethod
    def from_dict(cls, obj: dict):
        return cls(indices=obj.get("ids"), c=obj.get("c"))
    
    def evaluate(self, configuration: Dict[int, int]) -> float:
        """Given a variable configuration, evaluate the value of the term.
        :param configuration:
            The dictionary of variable ids to their assigned value
        """
        try:
            multiplier = (
                np.prod([configuration[i] for i in self.ids])
                if len(self.ids) > 0
                else 1.0
            )
        except KeyError:
            print(
                "Error - variable id found in term {0}, \
                    but not found in the supplied configuration.".format(
                    self.ids
                )
            )
            raise

        return multiplier * self.c
    
    def reduce_by_variable_state(
        self, fixed_variables: Dict[int, int]
    ) -> Optional[Term]:
        """Given some fixed variable states,
            transform the existing term into new term.
        Returns None if the new term is effectively 0
        :param fixed_variables:
            The dictionary of variable ids and their fixed state
        """
        new_ids = []
        new_c = self.c

        for i in self.ids:
            if i not in fixed_variables:
                new_ids.append(i)
            else:
                new_c *= fixed_variables[i]
                if new_c == 0:
                    return None

        return Term(indices=new_ids, c=new_c)


class GroupType(str, Enum):
    combination = "na"
    squared_linear_combination = "slc"

class SlcTerm(TermBase):
    """
    Squared Linear Combination term class.
    """
    def __init__(
        self,
        terms: List[Term],
        c: Optional[WArray] = 1.0,
    ):
        TermBase.__init__(self, c=c)
        self.terms = terms
        self.validate()
    
    def validate(self):
        """
        Check for and raise errors in Squared Linear Combination formulation.
        """
        # Check linearity of terms and that like terms are combined
        seen = set()
        for term in self.terms:
            if len(term.ids) > 1:
                # Nonlinear term
                raise ValueError("Error - terms must be linear in type SlcTerm")
            elif len(term.ids) == 1:
                # Linear term
                id = term.ids[0]
            else:
                # Constant term
                id = -1
            if id in seen:
                raise ValueError("Error - like terms must be combined in type SlcTerm")
            else:
                seen.add(id)


    def to_dict(self):
        """
        Return dictionary format of SlcTerm for solver input
        """
        return {
            'c': self.c,
            'terms': [monomial_term.to_dict() for monomial_term in self.terms],
        }
    
    @classmethod
    def from_dict(cls, obj: dict):
        """
        Create SlcTerm from dictionary with keys "terms" and "c"
        """
        try:
            terms = [Term.from_dict(term_dict) for term_dict in obj["terms"]]
        except:
            print(
                "Error - grouped list of terms missing or errant for squared linear combination type."
            )
            raise
        return cls(terms=terms, c=obj["c"])
    
    def evaluate(self, configuration: Dict[int, int]) -> float:
        """Given a variable configuration, evaluate the value of the slc term.
        :param configuration:
            Dictionary in which each key is a variable id;
            each value is the variable assignment (usually -1, 0, or 1)
        """
        combination_eval = 0.0
        for term in self.terms:
            combination_eval += term.evaluate(configuration)
        eval = self.c * combination_eval**2
        
        return eval
    
    def reduce_by_variable_state(
        self,
        fixed_variables: Dict[int, int]
    ) -> Optional[TermBase]:
        """Given some fixed variable states,
            transform the existing grouped term into new term.
        Returns None if the new term is effectively 0
        :param fixed_variables:
            The dictionary of variable ids and their fixed state
        """
        new_terms_dict = dict()
        for monomial in self.terms:
            new_monomial = monomial.reduce_by_variable_state(fixed_variables)
            if new_monomial:
                ids = tuple(sorted(new_monomial.ids))
                try:
                    new_terms_dict[ids] += new_monomial.c
                except:
                    new_terms_dict[ids] = new_monomial.c
        new_terms = [Term(indices=list(ids), c=c) for ids,c in new_terms_dict.items()]
        if len(new_terms) == 0:
            return None

        # Slc simplifications when new_terms has a single constant element
        # such as simplifying an SLC term consisting of a single variable
        if len(new_terms) == 1:
            term = new_terms[0]
            if len(term.ids) == 0:
                # Simplify SLC term consisting of a single constant
                # For example, C(k)^2 as a GroupedTerm to Ck^2 as a Term
                return Term(indices=[], c=self.c * term.c**2)

        return SlcTerm(terms=new_terms, c=self.c)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: problem.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='problem.proto',
  package='AzureQuantum',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\rproblem.proto\x12\x0c\x41zureQuantum\x1a\x1cgoogle/protobuf/struct.proto\"\xc7\x03\n\x07Problem\x12\x39\n\rcost_function\x18\x01 \x01(\x0b\x32\".AzureQuantum.Problem.CostFunction\x12)\n\x08metadata\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x1a\x1e\n\x04Term\x12\t\n\x01\x63\x18\x01 \x01(\x01\x12\x0b\n\x03ids\x18\x02 \x03(\x03\x1a\xf7\x01\n\x0c\x43ostFunction\x12/\n\x04type\x18\x01 \x01(\x0e\x32!.AzureQuantum.Problem.ProblemType\x12\x0f\n\x07version\x18\x02 \x01(\t\x12)\n\x05terms\x18\x03 \x03(\x0b\x32\x1a.AzureQuantum.Problem.Term\x12G\n\x0binit_config\x18\x04 \x03(\x0b\x32\x32.AzureQuantum.Problem.CostFunction.InitConfigEntry\x1a\x31\n\x0fInitConfigEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\"<\n\x0bProblemType\x12\x08\n\x04PUBO\x10\x00\x12\t\n\x05ISING\x10\x01\x12\n\n\x06MAXSAT\x10\x02\x12\x0c\n\x08SOFTSPIN\x10\x03\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])



_PROBLEM_PROBLEMTYPE = _descriptor.EnumDescriptor(
  name='ProblemType',
  full_name='AzureQuantum.Problem.ProblemType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PUBO', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ISING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MAXSAT', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SOFTSPIN', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=457,
  serialized_end=517,
)
_sym_db.RegisterEnumDescriptor(_PROBLEM_PROBLEMTYPE)


_PROBLEM_TERM = _descriptor.Descriptor(
  name='Term',
  full_name='AzureQuantum.Problem.Term',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='c', full_name='AzureQuantum.Problem.Term.c', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ids', full_name='AzureQuantum.Problem.Term.ids', index=1,
      number=2, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=175,
  serialized_end=205,
)

_PROBLEM_COSTFUNCTION_INITCONFIGENTRY = _descriptor.Descriptor(
  name='InitConfigEntry',
  full_name='AzureQuantum.Problem.CostFunction.InitConfigEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='AzureQuantum.Problem.CostFunction.InitConfigEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='AzureQuantum.Problem.CostFunction.InitConfigEntry.value', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=406,
  serialized_end=455,
)

_PROBLEM_COSTFUNCTION = _descriptor.Descriptor(
  name='CostFunction',
  full_name='AzureQuantum.Problem.CostFunction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='AzureQuantum.Problem.CostFunction.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='AzureQuantum.Problem.CostFunction.version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='terms', full_name='AzureQuantum.Problem.CostFunction.terms', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='init_config', full_name='AzureQuantum.Problem.CostFunction.init_config', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PROBLEM_COSTFUNCTION_INITCONFIGENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=208,
  serialized_end=455,
)

_PROBLEM = _descriptor.Descriptor(
  name='Problem',
  full_name='AzureQuantum.Problem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='cost_function', full_name='AzureQuantum.Problem.cost_function', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='AzureQuantum.Problem.metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PROBLEM_TERM, _PROBLEM_COSTFUNCTION, ],
  enum_types=[
    _PROBLEM_PROBLEMTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=517,
)

_PROBLEM_TERM.containing_type = _PROBLEM
_PROBLEM_COSTFUNCTION_INITCONFIGENTRY.containing_type = _PROBLEM_COSTFUNCTION
_PROBLEM_COSTFUNCTION.fields_by_name['type'].enum_type = _PROBLEM_PROBLEMTYPE
_PROBLEM_COSTFUNCTION.fields_by_name['terms'].message_type = _PROBLEM_TERM
_PROBLEM_COSTFUNCTION.fields_by_name['init_config'].message_type = _PROBLEM_COSTFUNCTION_INITCONFIGENTRY
_PROBLEM_COSTFUNCTION.containing_type = _PROBLEM
_PROBLEM.fields_by_name['cost_function'].message_type = _PROBLEM_COSTFUNCTION
_PROBLEM.fields_by_name['metadata'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_PROBLEM_PROBLEMTYPE.containing_type = _PROBLEM
DESCRIPTOR.message_types_by_name['Problem'] = _PROBLEM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Problem = _reflection.GeneratedProtocolMessageType('Problem', (_message.Message,), {

  'Term' : _reflection.GeneratedProtocolMessageType('Term', (_message.Message,), {
    'DESCRIPTOR' : _PROBLEM_TERM,
    '__module__' : 'problem_pb2'
    # @@protoc_insertion_point(class_scope:AzureQuantum.Problem.Term)
    })
  ,

  'CostFunction' : _reflection.GeneratedProtocolMessageType('CostFunction', (_message.Message,), {

    'InitConfigEntry' : _reflection.GeneratedProtocolMessageType('InitConfigEntry', (_message.Message,), {
      'DESCRIPTOR' : _PROBLEM_COSTFUNCTION_INITCONFIGENTRY,
      '__module__' : 'problem_pb2'
      # @@protoc_insertion_point(class_scope:AzureQuantum.Problem.CostFunction.InitConfigEntry)
      })
    ,
    'DESCRIPTOR' : _PROBLEM_COSTFUNCTION,
    '__module__' : 'problem_pb2'
    # @@protoc_insertion_point(class_scope:AzureQuantum.Problem.CostFunction)
    })
  ,
  'DESCRIPTOR' : _PROBLEM,
  '__module__' : 'problem_pb2'
  # @@protoc_insertion_point(class_scope:AzureQuantum.Problem)
  })
_sym_db.RegisterMessage(Problem)
_sym_db.RegisterMessage(Problem.Term)
_sym_db.RegisterMessage(Problem.CostFunction)
_sym_db.RegisterMessage(Problem.CostFunction.InitConfigEntry)


_PROBLEM_COSTFUNCTION_INITCONFIGENTRY._options = None
# @@protoc_insertion_point(module_scope)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

import abc
from typing import TYPE_CHECKING, Union

from azure.quantum._client.models import ItemDetails, ItemType, SessionDetails, JobDetails

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace

__all__ = ["WorkspaceItem"]

class WorkspaceItem(abc.ABC):
    """

    :param workspace: Workspace instance to submit job to
    :type workspace: Workspace
    :param item_details: Item details model,
            contains item ID, name and other details
    :type item_details: ItemDetails
    """

    def __init__(self, workspace: "Workspace", details: ItemDetails, **kwargs):
        self._workspace = workspace
        self._details = details
        self._id = details.id
        self._item_type = details.item_type

    @property
    def workspace(self) -> "Workspace":
        return self._workspace

    @property
    def details(self) -> Union[SessionDetails, JobDetails]:
        return self._details

    @property
    def id(self) -> str:
        return self._id

    @property
    def item_type(self) -> ItemType:
        return self._item_type



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from typing import TYPE_CHECKING, Any, Dict, Protocol, runtime_checkable
import io
import json
import abc

from azure.quantum._client.models import TargetStatus, SessionDetails
from azure.quantum._client.models._enums import SessionJobFailurePolicy
from azure.quantum.job.job import Job
from azure.quantum.job.session import Session, SessionHost
from azure.quantum.job.base_job import ContentType
if TYPE_CHECKING:
    from azure.quantum import Workspace


@runtime_checkable
class QirRepresentable(Protocol):
    _name: str

    @abc.abstractmethod
    def _repr_qir_(self, **kwargs: Any) -> bytes:
        raise NotImplementedError


class Target(abc.ABC, SessionHost):
    """Azure Quantum Target."""
    # Target IDs that are compatible with this Target class.
    # This variable is used by TargetFactory. To set the default
    # target class for a given provider, specify the
    # default_targets constructor argument.
    target_names = ()

    def __init__(
        self,
        workspace: "Workspace",
        name: str,
        input_data_format: str = "",
        output_data_format: str = "",
        capability: str = "",
        provider_id: str = "",
        content_type: ContentType = ContentType.json,
        encoding: str = "",
        average_queue_time: float = None,
        current_availability: str = ""
    ):
        if not provider_id and "." in name:
            provider_id = name.split(".")[0]
        self.workspace = workspace
        self.name = name
        self.input_data_format = input_data_format
        self.output_data_format = output_data_format
        self.capability = capability
        self.provider_id = provider_id
        self.content_type = content_type
        self.encoding = encoding
        self._average_queue_time = average_queue_time
        self._current_availability = current_availability

    def __repr__(self):
        return f"<Target name=\"{self.name}\", \
avg. queue time={self._average_queue_time} s, {self._current_availability}>"

    @classmethod
    def from_target_status(
        cls, workspace: "Workspace", status: TargetStatus, **kwargs
    ):
        """Create a Target instance from a given workspace and target status.

        :param workspace: Associated workspace
        :type workspace: Workspace
        :param status: Target status with availability and current queue time
        :type status: TargetStatus
        :return: Target instance
        :rtype: Target
        """
        return cls(
            workspace=workspace,
            name=status.id,
            average_queue_time=status.average_queue_time,
            current_availability=status.current_availability,
            **kwargs
        )
    
    def refresh(self):
        """Update the target availability and queue time"""
        targets = self.workspace._get_target_status(self.name, self.provider_id)
        if len(targets) > 0:
            _, target_status = targets[0]
            self._current_availability = target_status.current_availability
            self._average_queue_time = target_status.average_queue_time
        else:
            raise ValueError(
                f"Cannot refresh the Target status: \
target '{self.name}' of provider '{self.provider_id}' not found."
            )

    @property
    def current_availability(self):
        return self._current_availability
    
    @property
    def average_queue_time(self):
        return self._average_queue_time

    @staticmethod
    def _encode_input_data(data: Any) -> bytes:
        """Encode input data to bytes.
        If the data is already in bytes format, return it.

        :param data: Input data
        :type data: Any
        :return: Encoded input data
        :rtype: bytes
        """
        if isinstance(data, bytes):
            return data
        else:
            stream = io.BytesIO()
            if isinstance(data, dict):
                data = json.dumps(data)
            stream.write(data.encode())
            return stream.getvalue()

    def submit(
        self,
        input_data: Any,
        name: str = "azure-quantum-job",
        input_params: Dict[str, Any] = None,
        **kwargs
    ) -> Job:
        """Submit input data and return Job.

        Provide input_data_format, output_data_format and content_type
        keyword arguments to override default values.

        :param input_data: Input data
        :type input_data: Any
        :param name: Job name
        :type name: str
        :param input_params: Input parameters
        :type input_params: Dict[str, Any]
        :return: Azure Quantum job
        :rtype: Job
        """

        input_params = input_params or {}
        input_data_format = None
        output_data_format = None
        content_type = None

        # If the input_data is `QirRepresentable`
        # we need to convert it to QIR bitcode and set the necessary parameters for a QIR job.
        if input_data and isinstance(input_data, QirRepresentable):
            input_data_format = kwargs.pop("input_data_format", "qir.v1")
            output_data_format = kwargs.pop("output_data_format", "microsoft.quantum-results.v1")
            content_type = kwargs.pop("content_type", "qir.v1")

            def _get_entrypoint(input_data):
                # TODO: this method should be part of QirRepresentable protocol
                # and will later move to the QSharpCallable class in the qsharp package 
                import re
                method_name = re.search("(?:^|\\.)([^.]*)$", input_data._name).group(1)
                return f'ENTRYPOINT__{method_name}'

            input_params["entryPoint"] = input_params.get("entryPoint", _get_entrypoint(input_data))
            input_params["arguments"] = input_params.get("arguments", [])
            targetCapability = input_params.get("targetCapability", kwargs.pop("target_capability", self.capability))
            if targetCapability:
                input_params["targetCapability"] = targetCapability
            input_data = input_data._repr_qir_(target=self.name, target_capability=targetCapability)
        else:
            input_data_format = kwargs.pop("input_data_format", self.input_data_format)
            output_data_format = kwargs.pop("output_data_format", self.output_data_format)
            content_type = kwargs.pop("content_type", self.content_type)

        encoding = kwargs.pop("encoding", self.encoding)
        blob = self._encode_input_data(data=input_data)
        return Job.from_input_data(
            workspace=self.workspace,
            name=name,
            target=self.name,
            input_data=blob,
            content_type=content_type,
            encoding=encoding,
            provider_id=self.provider_id,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            input_params=input_params,
            session_id=self.get_latest_session_id(),
            **kwargs
        )

    def supports_protobuf(self):
        """
        Return whether or not the Solver class supports protobuf serialization.
        This should be overridden by Solver subclasses which do support protobuf.
        """
        return False
    
    def estimate_cost(
        self,
        input_data: Any,
        input_params: Dict[str, Any] = None
    ):
        return NotImplementedError("Price estimation is not implemented yet for this target.")

    def _get_azure_workspace(self) -> "Workspace":
        return self.workspace

    def _get_azure_target_id(self) -> str:
        return self.name



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from typing import Any, Dict

from azure.quantum.aio.target.target import Target
from azure.quantum.aio.job.job import Job
from azure.quantum.target import IonQ as SyncIonQ


class IonQ(Target, SyncIonQ):
    """IonQ target."""

    async def submit(
        self,
        circuit: Dict[str, Any],
        name: str = "ionq-job",
        num_shots: int = None,
        input_params: Dict[str, Any] = None,
        **kwargs
    ) -> Job:
        """Submit an IonQ circuit (JSON format)

        :param circuit: Quantum circuit in IonQ JSON format
        :type circuit: Dict[str, Any]
        :param name: Job name
        :type name: str
        :param num_shots: Number of shots, defaults to None
        :type num_shots: int
        :param input_params: Optional input params dict
        :type input_params: Dict[str, Any]
        :return: Azure Quantum job
        :rtype: Job
        """
        if input_params is None:
            input_params = {}
        if num_shots is not None:
            input_params = input_params.copy()
            input_params["shots"] = num_shots

        return await super().submit(
            input_data=circuit,
            name=name,
            input_params=input_params,
            **kwargs
        )



try:
    from qiskit.result import Result
    import markdown
    import copy
    from qiskit.result.models import ExperimentResult
    from qiskit.qobj import QobjHeader
except ImportError:
    raise ImportError(
        "Missing optional 'qiskit' dependencies. \
To install run: pip install azure-quantum[qiskit]"
    )

class ResourceEstimatorBatchResult(Result):
    """
    A customized result for a resource estimation batching job.
    """

    MAX_DEFAULT_ITEMS_IN_TABLE = 5

    def __init__(
            self,
            **kwargs
    ) -> None:
        super().__init__(**kwargs)

        # Creates individual result items for each item in the batch results
        self.result_items = [ResourceEstimatorResult(**{**kwargs, "results": [result]}) for result in self.results]

    @classmethod
    def from_dict(cls, data):
        """
        Overrides default behavior from Result.from_dict to account for batching
        data.
        """
        in_data = copy.copy(data)

        # We know that we have exactly one result here
        result = in_data.pop("results")[0]
        # All resource estimation results are in the data field, which we are going to flatten over all results
        data = result["data"]
        in_data["results"] = [ExperimentResult.from_dict({**result, "data": d}) for d in data]
        if "header" in in_data:
            in_data["header"] = QobjHeader.from_dict(in_data.pop("header"))
        return cls(**in_data)

    def __len__(self):
        """
        Returns the number of items in the batching job.
        """
        return len(self.results)

    def _repr_html_(self):
        """
        HTML representation of the result object; dispatches based on job
        success.
        """
        return self._repr_html_success_() if self.success else self._repr_html_error_()

    def _repr_html_success_(self):
        num_items = len(self.results)
        if num_items > self.MAX_DEFAULT_ITEMS_IN_TABLE:
            html = f"<p><b>Info:</b> <i>The overview table is cut off after {self.MAX_DEFAULT_ITEMS_IN_TABLE} items.  If you want to see all items, suffix the result variable with <code>[0:{num_items}]</code></i></p>"
            return html + _batch_result_html_table(self, range(self.MAX_DEFAULT_ITEMS_IN_TABLE))
        else:
            return _batch_result_html_table(self, range(len(self.results)))

    def _repr_html_error_(self):
        html = "<b>Job not successful</b>"
        if self.error_data is not None:
            html += f"""<br>Error code: {self.error_data['code']}
                <br>Error message: {self.error_data['message']}"""
        return html

    def plot(self, **kwargs):
        """
        Plots all result items in a space time plot, where the x-axis shows
        total runtime, and the y-axis shows total number of physical qubits.
        Both axes are in log-scale.

        Attributes:
            labels (list): List of labels for the legend.
        """
        import matplotlib.pyplot as plt

        labels = kwargs.pop("labels", [])

        [xs, ys] = zip(*[(self.data(i)['physicalCounts']['runtime'], self.data(i)['physicalCounts']['physicalQubits']) for i in range(len(self))])

        _ = plt.figure(figsize=(15, 8))

        plt.ylabel('Physical qubits')
        plt.xlabel('Runtime')
        plt.loglog()
        for i, (x, y) in enumerate(zip(xs, ys)):
            if isinstance(labels, list) and i < len(labels):
                label = labels[i]
            else:
                label = str(i)
            plt.scatter(x=[x], y=[y], label=label, marker="os+x"[i % 4])

        nsec = 1
        usec = 1e3 * nsec
        msec = 1e3 * usec
        sec = 1e3 * msec
        min = 60 * sec
        hour = 60 * min
        day = 24 * hour
        week = 7 * day
        month = 31 * day
        year = 365 * month
        decade = 10 * year
        century = 10 * decade

        time_units = [nsec, usec, msec, sec, min, hour, day, week, month, year, decade, century]
        time_labels = ["1 ns", "1 µs", "1 ms", "1 s", "1 min", "1 hour", "1 day", "1 week", "1 month", "1 year", "1 decade", "1 century"]

        cutoff = next((i for i, x in enumerate(time_units) if x > max(xs)), len(time_units) - 1) + 1

        plt.xticks(time_units[0:cutoff], time_labels[0:cutoff], rotation=90)
        plt.legend(loc="upper left")
        plt.show()

    def __getitem__(self, key):
        if isinstance(key, slice):
            from IPython.display import display, HTML
            display(HTML(_batch_result_html_table(self, range(len(self.results))[key])))
        else:
            return self.result_items[key]

class ResourceEstimatorResult(Result):
    """
    A customized result for a resource estimation job.
    """

    def __init__(
        self,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)

        # We also create a summary view based on a different result type.
        self.summary = ResourceEstimatorResultSummary(**kwargs)

    def _repr_html_(self):
        """
        HTML representation of the result object; dispatches based on job
        success.
        """
        return self._repr_html_success_() if self.success else self._repr_html_error_()

    def _repr_html_success_(self):
        html = ""

        md = markdown.Markdown(extensions=['mdx_math'])
        for group in self.data()['reportData']['groups']:
            html += f"""
                <details {"open" if group['alwaysVisible'] else ""}>
                    <summary style="display:list-item">
                        <strong>{group['title']}</strong>
                    </summary>
                    <table>"""
            for entry in group['entries']:
                val = self.data()
                for key in entry['path'].split("/"):
                    val = val[key]
                explanation = md.convert(entry["explanation"])
                html += f"""
                    <tr>
                        <td style="font-weight: bold; vertical-align: top; white-space: nowrap">{entry['label']}</td>
                        <td style="vertical-align: top; white-space: nowrap">{val}</td>
                        <td style="text-align: left">
                            <strong>{entry["description"]}</strong>
                            <hr style="margin-top: 2px; margin-bottom: 0px; border-top: solid 1px black" />
                            {explanation}
                        </td>
                    </tr>
                """
            html += "</table></details>"

        html += f"<details><summary style=\"display:list-item\"><strong>Assumptions</strong></summary><ul>"
        for assumption in self.data()['reportData']['assumptions']:
            html += f"<li>{md.convert(assumption)}</li>"
        html += "</ul></details>"

        return html

    def _repr_html_error_(self):
        html = "<b>Job not successful</b>"
        if self.error_data is not None:
            html += f"""<br>Error code: {self.error_data['code']}
                <br>Error message: {self.error_data['message']}"""
        return html

class ResourceEstimatorResultSummary(Result):
    def __init__(
        self,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)

    def _repr_html_(self):
        return self._repr_html_success_() if self.success else self._repr_html_error_()

    def _repr_html_success_(self):
        html = """
            <style>
                .aqre-tooltip {
                    position: relative;
                    border-bottom: 1px dotted black;
                }

                .aqre-tooltip .aqre-tooltiptext {
                    font-weight: normal;
                    visibility: hidden;
                    width: 600px;
                    background-color: #e0e0e0;
                    color: black;
                    text-align: center;
                    border-radius: 6px;
                    padding: 5px 5px;
                    position: absolute;
                    z-index: 1;
                    top: 150%;
                    left: 50%;
                    margin-left: -200px;
                    border: solid 1px black;
                }

                .aqre-tooltip .aqre-tooltiptext::after {
                    content: "";
                    position: absolute;
                    bottom: 100%;
                    left: 50%;
                    margin-left: -5px;
                    border-width: 5px;
                    border-style: solid;
                    border-color: transparent transparent black transparent;
                }

                .aqre-tooltip:hover .aqre-tooltiptext {
                    visibility: visible;
                }
            </style>"""

        md = markdown.Markdown(extensions=['mdx_math'])
        for group in self.data()['reportData']['groups']:
            html += f"""
                <details {"open" if group['alwaysVisible'] else ""}>
                    <summary style="display:list-item">
                        <strong>{group['title']}</strong>
                    </summary>
                    <table>"""
            for entry in group['entries']:
                val = self.data()
                for key in entry['path'].split("/"):
                    val = val[key]
                explanation = md.convert(entry["explanation"])
                html += f"""
                    <tr class="aqre-tooltip">
                        <td style="font-weight: bold"><span class="aqre-tooltiptext">{explanation}</span>{entry['label']}</td>
                        <td>{val}</td>
                        <td style="text-align: left">{entry["description"]}</td>
                    </tr>
                """
            html += "</table></details>"

        html += f"<details><summary style=\'display:list-item\'><strong>Assumptions</strong></summary><ul>"
        for assumption in self.data()['reportData']['assumptions']:
            html += f"<li>{md.convert(assumption)}</li>"
        html += "</ul></details>"

        return html

    def _repr_html_error_(self):
        html = "<b>Job not successful</b>"
        if self.error_data is not None:
            html += f"""<br>Error code: {self.error_data['code']}
                <br>Error message: {self.error_data['message']}"""
        return html

def _batch_result_html_table(result, indices):
    html = ""

    md = markdown.Markdown(extensions=['mdx_math'])

    item_headers = "".join(f"<th>{i}</th>" for i in indices)

    for group_index, group in enumerate(result.data(0)['reportData']['groups']):
        html += f"""
            <details {"open" if group['alwaysVisible'] else ""}>
                <summary style="display:list-item">
                    <strong>{group['title']}</strong>
                </summary>
                <table>
                    <thead><tr><th>Item</th>{item_headers}</tr></thead>"""

        visited_entries = set()

        for entry in [entry for index in indices for entry in result.data(index)['reportData']['groups'][group_index]['entries']]:
            label = entry['label']
            if label in visited_entries:
                continue
            visited_entries.add(label)

            html += f"""
                <tr>
                    <td style="font-weight: bold; vertical-align: top; white-space: nowrap">{label}</td>
            """

            for index in indices:
                val = result.data(index)
                for key in entry['path'].split("/"):
                    if key in val:
                        val = val[key]
                    else:
                        val = "N/A"
                        break
                html += f"""
                        <td style="vertical-align: top; white-space: nowrap">{val}</td>
                """

            html += """
                </tr>
            """
        html += "</table></details>"

    html += f"<details><summary style=\"display:list-item\"><strong>Assumptions</strong></summary><ul>"
    for assumption in result.data(0)['reportData']['assumptions']:
        html += f"<li>{md.convert(assumption)}</li>"
    html += "</ul></details>"

    return html



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from __future__ import annotations
import logging
import io
import gzip
import json
import numpy
import os
import tarfile


from typing import List, Tuple, Union, Dict, Optional, TYPE_CHECKING
from enum import Enum
from azure.quantum.optimization import TermBase, Term, GroupType, SlcTerm
from azure.quantum.storage import (
    ContainerClient,
    download_blob,
    BlobClient,
    download_blob_metadata,
    download_blob_properties
)
from azure.quantum.job.base_job import ContentType
from azure.quantum.job.job import Job
from azure.quantum.target.target import Target
from azure.quantum.serialization import ProtoProblem
from google.protobuf import struct_pb2

logger = logging.getLogger(__name__)

__all__ = ["Problem", "ProblemType"]

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace

class ProblemType(str, Enum):
    pubo = 0
    ising = 1
    pubo_grouped = 2
    ising_grouped = 3

proto_types = {
    ProblemType.ising: ProtoProblem.ProblemType.ISING,
    ProblemType.pubo: ProtoProblem.ProblemType.PUBO, 
}


class Problem:
    """Problem to submit to the service.

    :param name: Problem name
    :type name: str
        When passing None, it will default to "Optimization problem"
    :param terms: Problem terms, depending on solver.
        Defaults to None
    :type terms: Optional[List[TermBase]], optional
    :param init_config: Optional configuration details, depending on solver.
        Defaults to None
    :type init_config: Optional[Dict[str,int]], optional
    :param problem_type: Problem type (ProblemType.pubo or
        ProblemType.ising), defaults to ProblemType.ising
    :type problem_type: ProblemType, optional
    :param content_type: Content type, eg: application/json. application/x-protobuf. Default is application/json
    :type content_type: ContentType, optional
    """

    def __init__(
        self,
        name: str,
        terms: Optional[List[TermBase]] = None,
        init_config: Optional[Dict[str, int]] = None,
        problem_type: ProblemType = ProblemType.ising,
        content_type: Optional[ContentType] = ContentType.json 
    ):
        self.name = name or "Optimization problem"
        self.problem_type = problem_type
        self.init_config = init_config
        self.uploaded_blob_uri = None
        self.uploaded_blob_params = None
        self.content_type = content_type

        # each type of term has its own section for quicker serialization
        self.terms = []
        self.terms_slc = []

        # set the terms
        if terms:
            for term in terms:
                if isinstance(term, SlcTerm):
                    self.terms_slc.append(term)
                else:
                    self.terms.append(term)

        self.check_for_grouped_term()

    """
    Constant thresholds used to determine if a problem is "large".
    """
    NUM_VARIABLES_LARGE = 2500
    NUM_TERMS_LARGE = 1e6


    def serialize(self) -> Union[str, list]:
        """Wrapper function for serialzing. It may serialize to json or protobuf
        """ 
        if (self.content_type == ContentType.protobuf and len(self.terms_slc) == 0):
            return self.to_proto()
        else:
            return self.to_json()

    def to_json(self) -> str:
        """Serializes the problem to a JSON string"""
        result = {
            "metadata": {
                "name": self.name,
            },
            "cost_function": {
                "version": "1.1" if self.init_config else "1.0",
                "type": self.problem_type.name,
                "terms": [term.to_dict() for term in self.terms]
            }
        }
        if len(self.terms_slc) > 0:
            result["cost_function"]["terms_slc"] = [term.to_dict() for term in self.terms_slc]

        if self.init_config:
            result["cost_function"]["initial_configuration"] = self.init_config

        return json.dumps(result)
    
    def to_proto(self) -> list:
        """Serializes a problem to a list serialized protobuf messages
        Every problem is built into a series of protobuf messages 
        each with 1000 terms and added to a list of proto messages.
        In this version only ising and pubo are supported. The grouped terms will be added in the
        subsquent PR.
        Every message in the list is a byte string
        """
        proto_messages = []
        msg_count = 0
        terms_remaining = len(self.terms)
        terms_read = 0
        while terms_remaining > 0:   
            proto_problem = ProtoProblem()
            cost_function = proto_problem.cost_function
            metadata = proto_problem.metadata
            if msg_count == 0:
                cost_function.version = "1.0"
                cost_function.type = proto_types[self.problem_type]
                metadata["name"] = self.name
            # add 1000 terms per proto message
            if terms_remaining - 1000 > 0: 
                for i in range (terms_read,terms_read + 1000):
                    term = cost_function.terms.add()
                    term.c = self.terms[i].c
                    for j in range (len(self.terms[i].ids)):
                        term.ids.append(self.terms[i].ids[j])
            else:
                # add the remaining terms to the last message
                for i in range(terms_remaining):
                    term = cost_function.terms.add()
                    term.c = self.terms[i].c
                    for j in range (len(self.terms[i].ids)):
                        term.ids.append(self.terms[i].ids[j])
            msg_count += 1
            terms_remaining -= 1000
            terms_read += 1000
            proto_messages.append(proto_problem.SerializeToString())
        return proto_messages
    
    def compress_protobuf(self, 
    proto_messages: List[str] ) -> bytes:
    # Write to a series of files to folder and compress
        data = io.BytesIO()
        file_name_prefix = "gzipinputfile_pb"
        file_count = 0
        with tarfile.open(fileobj = data, mode = 'w:gz') as tar:
            for msg in proto_messages:
                file_name = file_name_prefix+"_"+str(file_count)+".pb"
                info = tarfile.TarInfo(name=file_name)
                info.size = len(msg)
                msg_data = io.BytesIO(msg)
                tar.addfile(info, msg_data)
                file_count += 1
        return data.getvalue() 

    
    @classmethod
    def from_json(
            cls, 
            input_problem: str, 
            name: Optional[str] = None
        ) -> Problem:
        """Deserializes the problem from a
        json serialized with Problem.serialize()

        :param input_problem:
            the json string to be deserialized to a `Problem` instance
        :type problem_msgs: str
        :param
        :param name: 
            The name of the problem is optional, since it will try 
            to read the serialized name from the json payload.
            If this parameter is not empty, it will use it as the
            problem name ignoring the serialized value.
        :type name: Optional[str]
        """
        result = json.loads(input_problem)

        if name is None:
            metadata = result.get("metadata")
            if metadata is not None:
                name = metadata.get("name")

        terms = [Term.from_dict(t) for t in result["cost_function"]["terms"]] if "terms" in result["cost_function"] else []
        if "terms_slc" in result["cost_function"]:
            terms.append([SlcTerm.from_dict(t) for t in result["cost_function"]["terms_slc"]])

        problem = cls(
            name=name,
            terms=terms,
            problem_type=ProblemType[result["cost_function"]["type"]],
        )

        if "initial_configuration" in result["cost_function"]:
            problem.init_config = result["cost_function"]["initial_configuration"]

        return problem
    
    @classmethod
    def from_proto(
        cls,
        input_problem: list,
        name: Optional[str] = None
    ) -> Problem:
        """Deserializes the problem from a
        protobuf messages serialized with Problem.serialize()

        :param input_problem:
            the list of protobuf messages to be deserialized to a `Problem` instance
        :type input_problem: list
        :param
        :param name: 
            The name of the problem is optional, since it will try 
            to read the serialized name from the json payload.
            If this parameter is not empty, it will use it as the
            problem name ignoring the serialized value.
        :type name: Optional[str]
        """
        msg_count = 0

        problem = cls(
            name = name
        )

        for msg in input_problem:
            proto_problem = ProtoProblem()
            proto_problem.ParseFromString(msg)
            if msg_count == 0:
                for qdk_type, proto_type in proto_types.items():
                    if proto_problem.cost_function.type == proto_type:
                        problem.problem_type = qdk_type
                metadata = proto_problem.metadata               
                if name is None:
                    name = metadata["name"]
                    problem.name = name
            for msg_term in proto_problem.cost_function.terms:
                term = Term(
                    c = msg_term.c
                )
                term.ids = []
                for msg_term_id in msg_term.ids:
                    term.ids.append(msg_term_id)
                problem.terms.append(term)
        
        return problem

    @classmethod
    def deserialize(
        cls, 
        input_problem: Union[str, list],
        name: Optional[str] = None, 
        content_type: Optional[ContentType] = None) -> Problem:
        """Deserializes the problem from a
        JSON string or protobuf messages serialized with Problem.serialize()
        Also used to deserialize the messages downloaded from the blob

        :param input_problem:
            The json string or the list of protobuf messages to be deserialized to a `Problem` instance
        :type input_problem: Union[str,list]
        :param
        :param name: 
            The name of the problem is optional, since it will try 
            to read the serialized name from the json payload.
            If this parameter is not empty, it will use it as the
            problem name ignoring the serialized value.
        :type name: Optional[str]
        :param content_type: The content type of the input problem data
        :type: Optional, ContentType
        """
        if content_type == ContentType.protobuf or type(input_problem) == list :
            return cls.from_proto(input_problem, name) 
        else :
            return cls.from_json(input_problem, name)

    def add_term(self, c: Union[int, float], indices: List[int]):
        """Adds a single monomial term to the `Problem` representation

        :param c: The cost or weight of this term
        :type c: int, float
        :param indices: The variable indices that are in this term
        :type indices: List[int]
        """
        self.terms.append(Term(indices=indices, c=c))
        self.uploaded_blob_uri = None

    def add_terms(
        self, 
        terms: List[Term],
        term_type: GroupType=GroupType.combination,
        c: Union[int, float]=1
    ):
        """Adds an optionally grouped list of monomial terms 
        to the `Problem` representation

        :param terms: The list of terms to add to the problem
        :param term_type: Type of grouped term being added
            If GroupType.combination, terms will be added as ungrouped monomials.
        :param c: Weight of grouped term, only applicable for grouped terms that support it.
        """
        if term_type is GroupType.combination:
            # Default type is a group of monomials
            self.terms += terms
        else:
            # Slc term
            self.terms_slc.append(SlcTerm(terms=terms, c=c))
            self.problem_type_to_grouped()
        self.uploaded_blob_uri = None
    
    def add_slc_term(
        self,
        terms: Union[List[Tuple[Union[int, float], Optional[int]]],
                     List[Term]],
        c: Union[int, float] = 1
    ):
        """Adds a squared linear combination term
        to the `Problem` representation. Helper function to construct terms list.
        
        :param terms: List of monomial terms, with each represented by a pair.
            The first entry represents the monomial term weight.
            The second entry is the monomial term variable index or None.
            Alternatively, a list of Term objects may be input.
        :param c: Weight of SLC term
        """
        if all(isinstance(term, Term) for term in terms):
            gterms = terms
        else:
            gterms = [Term([index], c=tc) if index is not None else Term([], c=tc)
                      for tc,index in terms]
        self.terms_slc.append(
            SlcTerm(gterms, c=c)
        )
        self.problem_type_to_grouped()
        self.uploaded_blob_uri = None

    def check_for_grouped_term(self):
        if len(self.terms_slc) != 0:
            self.problem_type_to_grouped()
    
    def problem_type_to_grouped(self):
        if self.problem_type == ProblemType.pubo:
            self.problem_type = ProblemType.pubo_grouped
        elif self.problem_type == ProblemType.ising:
            self.problem_type = ProblemType.ising_grouped
            
    def to_blob(self) -> bytes:
        """Convert problem data to a binary blob.

        :return: Blob data
        :rtype: bytes
        """
        input_problem = self.serialize()
        debug_input_string = input_problem if type(input_problem) is str else b''.join( input_problem).decode('latin-1')
        logger.debug("Input Problem: " + debug_input_string)
        data = io.BytesIO()
        if self.content_type == ContentType.protobuf:
          return self.compress_protobuf(input_problem)                   
        else:
            with gzip.GzipFile(fileobj=data, mode="w") as fo:
                fo.write(input_problem.encode())

            return data.getvalue()
    
    def _blob_name(self):
        import uuid
        return "{}-{}".format(self.name, uuid.uuid1())

    def upload(
        self,
        workspace: "Workspace",
        container_name: str = "qio-problems",
        blob_name: str = "inputData",
        container_uri: str = None,
    ):
        """Uploads an optimization problem instance to
        the cloud storage linked with the Workspace.

        :param workspace: interaction terms of the problem.
        :type workspace: Workspace
        :param container_name: Container name, defaults to "qio-problems"
        :type container_name: str, optional
        :param blob_name: Blob name, defaults to None
        :type blob_name: str, optional
        :param container_uri: Optional container URI
        :type container_uri: str
        :return: uri of the uploaded problem
        :rtype: str
        """
        blob_params = [workspace, container_name, blob_name]
        if self.uploaded_blob_uri and self.uploaded_blob_params == blob_params:
            return self.uploaded_blob_uri

        if blob_name is None:
            blob_name = self._blob_name()

        encoding = "gzip"
        content_type = self.content_type

        blob = self.to_blob()
        if container_uri is None:
            container_uri = workspace.get_container_uri(
                container_name=container_name
            )
        input_data_uri = Job.upload_input_data(
            input_data=blob,
            blob_name=blob_name,
            container_uri=container_uri,
            encoding=encoding,
            content_type= content_type
        )
        self.uploaded_blob_params = blob_params
        self.uploaded_blob_uri = input_data_uri
        return input_data_uri

    def set_fixed_variables(
        self, fixed_variables: Union[Dict[int, int], Dict[str, int]]
    ) -> Problem:
        """Transforms the current problem with a set of fixed
        variables and returns the new modified problem.
        The original Problem instance is untouched.

        :param fixed_variables:
            The dictionary of variable ids and their fixed state
        """
        if len(fixed_variables) == 0:
            raise RuntimeError(
                "Error: fixed_variables is empty - \
                please specify at least one fixed variable"
            )

        fixed_transformed = {
            int(k): fixed_variables[k] for k in fixed_variables
        }  # if ids are given in string form, convert them to int
        new_terms = []

        constant = 0
        for terms in [self.terms, self.terms_slc]:
            for term in terms:
                reduced_term = term.reduce_by_variable_state(fixed_transformed)
                if reduced_term:
                    if not isinstance(reduced_term, Term) or len(reduced_term.ids) > 0:
                        new_terms.append(reduced_term)
                    else:
                        # reduced to a constant term
                        constant += reduced_term.c

        if constant:
            new_terms.append(Term(c=constant, indices=[]))

        new_init_config = None
        if self.init_config:
            new_init_config = {
                k: self.init_config[k]
                for k in self.init_config
                if int(k) not in fixed_transformed
            }

        return Problem(
            self.name,
            terms=new_terms,
            init_config=new_init_config,
            problem_type=self.problem_type,
        )

    def _evaluate(self, configuration, term_list):
        total = 0
        if term_list:
            for term in term_list:
                total += term.evaluate(configuration)
        return total

    def evaluate(self, configuration: Union[Dict[int, int], Dict[str, int]]) -> float:
        """Given a configuration/variable assignment,
        return the cost function value of this problem.

        :param configuration: The dictionary of
         variable ids to their assigned value
        """
        configuration_transformed = {
            int(k): configuration[k] for k in configuration
        }  # if ids are given in string form, convert them to int

        total_cost = 0
        for terms in [self.terms, self.terms_slc]:
            total_cost += self._evaluate(configuration_transformed, terms)

        return total_cost

    def is_large(self) -> bool:
        """Determines if the current problem is large.
        "large" is an arbitrary threshold and can be easily changed.
        Based on usage data, we have defined a
        large problem to be NUM_VARIABLES_LARGE+
        variables AND NUM_TERMS_LARGE+ terms.
        """

        set_vars = set()
        total_term_count = 0
        for terms in [self.terms, self.terms_slc]:
            for term in terms:
                if isinstance(term, Term):
                    set_vars.update(term.ids)
                    total_term_count += 1
                elif isinstance(term, SlcTerm):
                    for subterm in term.terms:
                        set_vars.update(subterm.ids)
                        total_term_count += 1
        
        return (
            len(set_vars) >= Problem.NUM_VARIABLES_LARGE
            and total_term_count >= Problem.NUM_TERMS_LARGE
        )

    def download(self, workspace: "Workspace"):
        """Downloads the uploaded problem as an instance of `Problem`"""
        if not self.uploaded_blob_uri:
            raise Exception("Problem may not be downloaded before it is uploaded")
        blob_client = BlobClient.from_blob_url(self.uploaded_blob_uri)
        container_client = ContainerClient.from_container_url(
            workspace._get_linked_storage_sas_uri(blob_client.container_name)
        )
        blob_name = blob_client.blob_name
        blob = container_client.get_blob_client(blob_name)
        contents = download_blob(blob.url)
        blob_properties = download_blob_properties(blob.url)        
        content_type = blob_properties.content_type
        return Problem.deserialize(contents, self.name, content_type)

    def get_terms(self, id: int) -> List[TermBase]:
        """Given an index the function will return
        a list of terms with that index
        """
        terms = []
        if self.terms != [] or self.terms_slc != []:
            for all_terms in [self.terms, self.terms_slc]:
                for term in all_terms:
                    if isinstance(term, Term):
                        if id in term.ids:
                            terms.append(term)
                    elif isinstance(term, SlcTerm):
                        for subterm in term.terms:
                            if id in subterm.ids:
                                terms.append(term)
                                break
            return terms
        else:
            raise Exception(
                "There are currently no terms in this problem. \
                Please download the problem on the client or add terms to the \
                    problem to perform this operation"
            )

    def is_valid_npz(
        self,
        files: List[str],
        indices_column_names: List[str] = ["arr_0", "arr_1"],
        c_column_name: str = "arr_2",
    ) -> bool:
        """Determines if the supplied npz has expected column names.
        If none are supplied, checks default naming.
        Otherwise, it checks user-supplied naming.

        :param files: the associated column names of an npz file
        :type: List[str]
        :param indices_column_names: the names of the indices columns
        :type indices_column_names: List[str], optional
        :param c_column_name: the name of the coefficient column
        :type c_column_name: str, optional
        """

        all_columns = indices_column_names + [c_column_name]

        if len(files) != len(all_columns):
            return False

        if sorted(files) != sorted(all_columns):
            return False

        return True

    def terms_from_npz(
        self,
        file_path=str,
        indices_column_names: List[str] = ["arr_0", "arr_1"],
        c_column_name: str = "arr_2",
    ) -> List[Term]:
        """Reads a user supplied npz file and converts it to a list of `Term`.
        An NPZ file contains several arrays (or columns), which can specify
        the indices of a problem term, along with the coefficient.
        Default naming for these columns is used unless specified by the user.

        :param file_path: file path of the NPZ file to be converted
        :type file_name: str
        :param indices_column_names: the names of the indices columns
        :type indices_column_names: List[str], optional
        :param c_column_name: the name of the coefficient column
        :type c_column_name: str, optional
        """

        terms = []

        if os.path.isfile(file_path):
            problem_file = numpy.load(file_path)

            # Check the default or user-supplied columns are present
            if not (
                self.is_valid_npz(
                    problem_file.files, indices_column_names, c_column_name
                )
            ):
                raise Exception(
                    "Error in validating NPZ file. \
                    Please check that the names of the arrays match the default \
                        or user-supplied namings."
                )

            # For each of the indices column names, extract the associated indices column data
            problem_ids = [list(problem_file[id]) for id in indices_column_names]

            # Unpack the problem indices and zip with the coefficient column data
            # to produce problem terms
            for term in zip(*problem_ids, problem_file[c_column_name]):
                ids = list(term[:-1])
                indices = list(map(int, ids))

                c = float(term[-1])
                terms.append(Term(c=c, indices=indices))
            return terms
        else:
            raise Exception(
                "Unable to read NPZ file. \
                Please check the file path supplied is correct."
            )



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
from typing import TYPE_CHECKING, Any, Dict, Union

try:
    import cirq
    from cirq_ionq import Job as CirqIonqJob
    from cirq_ionq.results import QPUResult, SimulatorResult
except ImportError:
    raise ImportError(
        "Missing optional 'cirq_ionq' dependency. \
To install run: pip install azure-quantum[cirq]")

from azure.quantum.target import IonQ
from azure.quantum.cirq.targets.target import Target as CirqTarget

if TYPE_CHECKING:
    from azure.quantum import Workspace
    from azure.quantum.job import Job as AzureJob


class _IonQClient:
    """Thin wrapper around Workspace to support cirq_ionq.Job"""
    def __init__(self, workspace: "Workspace"):
        self._workspace = workspace
    
    @staticmethod
    def _to_ionq_status(status: str):
        from azure.quantum._client.models._enums import JobStatus
        _STATUS_DICT = {
            JobStatus.SUCCEEDED: 'completed',
            JobStatus.CANCELLED: 'canceled',
            JobStatus.FAILED: 'failed',
            JobStatus.EXECUTING: 'running',
            JobStatus.WAITING: 'ready'
        }
        return _STATUS_DICT.get(status, 'submitted')
    
    def _create_job_dict(self, azure_job: "AzureJob"):
        metadata = azure_job.details.input_params.copy()
        metadata.update(azure_job.details.metadata or {})

        if azure_job.has_completed():
            results = azure_job.get_results()
        else:
            results = None

        job_dict = {
            "id": azure_job.id,
            "name": azure_job.details.name,
            "status": self._to_ionq_status(azure_job.details.status),
            "data": results,
            "metadata": metadata,
            "target": azure_job.details.target,
            "qubits": metadata.get("qubits"),
            "failure": azure_job.details.error_data or ""
        }

        return job_dict
    
    def get_job(self, job_id: str):
        azure_job = self._workspace.get_job(job_id)
        return self._create_job_dict(azure_job)

    def cancel_job(self, job_id: str):
        azure_job = self._workspace.get_job(job_id)
        self._workspace.cancel_job(azure_job)

    def delete_job(self, job_id: str):
        azure_job = self._workspace.get_job(job_id)
        self._workspace.cancel_job(azure_job)


class IonQTarget(IonQ, CirqTarget):
    """Base class for interfacing with an IonQ backend in Azure Quantum"""

    def __init__(
        self,
        workspace: "Workspace",
        name: str,
        input_data_format: str = "ionq.circuit.v1",
        output_data_format: str = "ionq.quantum-results.v1",
        provider_id: str = "IonQ",
        content_type: str = "application/json",
        encoding: str = "",
        **kwargs
    ):
        self._client = _IonQClient(workspace)
        super().__init__(
            workspace=workspace,
            name=name,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            provider_id=provider_id,
            content_type=content_type,
            encoding=encoding,
            **kwargs
        )

    @staticmethod
    def _translate_cirq_circuit(circuit) -> Dict[str, Any]:
        """Translate Cirq circuit to IonQ JSON. If dependencies \
are not installed, throw error with installation instructions."""
        from cirq_ionq import Serializer
        return Serializer().serialize(circuit)

    def _to_cirq_job(self, azure_job: "AzureJob") -> "CirqIonqJob":
        """Convert Azure job to Cirq job"""
        job_dict = self._client._create_job_dict(azure_job)
        return CirqIonqJob(client=self._client, job_dict=job_dict)

    def estimate_cost(
        self,
        program: "cirq.Circuit",
        repetitions: int,
        price_1q: float = None,
        price_2q: float = None,
        min_price: float = None) -> float:
        """Estimate cost for running this program

        :param program: Cirq quantum program
        :type program: cirq.Circuit
        :param repetitions: Number of repetitions
        :type repetitions: int
        :param price_1q: The price of running a single-qubit gate.
        :type price_1q: float, optional
        :param price_2q: The price of running a double-qubit gate.
        :type price_2q: float, optional
        :param min_price: The minimum price for running a job.
        :type min_price: float, optional
        :return: Price estimate
        :rtype: float
        """
        serialized_program = self._translate_cirq_circuit(program)
        return super().estimate_cost(
            serialized_program.body,
            repetitions,
            price_1q=price_1q,
            price_2q=price_2q,
            min_price=min_price
        )

    def submit(
        self,
        program: "cirq.Circuit",
        name: str = "cirq-job",
        repetitions: int = 500,
        **kwargs
    ) -> "CirqIonqJob":
        """Submit a Cirq quantum circuit

        :param program: Cirq circuit
        :type program: cirq.Circuit
        :param name: Job name
        :type name: str
        :param repetitions: Number of shots, defaults to 
            provider default value
        :type repetitions: int
        :return: Azure Quantum job
        :rtype: Job
        """
        serialized_program = self._translate_cirq_circuit(program)
        metadata = serialized_program.metadata or {}
        metadata["qubits"] = serialized_program.body["qubits"]
        # Override metadata with value from kwargs
        metadata.update(kwargs.get("metadata", {}))
        azure_job = super().submit(
            circuit=serialized_program.body,
            name=name,
            num_shots=repetitions,
            metadata=metadata,
            **kwargs
        )
        # Get Job shim to process results using cirq.Result
        return self._to_cirq_job(azure_job=azure_job)
    
    @staticmethod
    def _to_cirq_result(
        result: Union[QPUResult, SimulatorResult],
        param_resolver: cirq.ParamResolverOrSimilarType = cirq.ParamResolver({}),
        seed: cirq.RANDOM_STATE_OR_SEED_LIKE = None,
    ) -> "cirq.Result":
        if isinstance(result, QPUResult):
            return result.to_cirq_result(params=cirq.ParamResolver(param_resolver))
        elif isinstance(result, SimulatorResult):
            return result.to_cirq_result(params=cirq.ParamResolver(param_resolver), seed=seed)

        raise ValueError("Result {result} not supported. \
            Expecting either a cirq_ionq.results.QPUResult \
                or cirq_ionq.results.SimulatorResult.")
 


##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
from collections import defaultdict
from typing import Any, Dict, List, Union
import numpy as np

try:
    from qiskit.providers import JobV1, JobStatus
    from qiskit.result import Result
except ImportError:
    raise ImportError(
        "Missing optional 'qiskit' dependencies. \
To install run: pip install azure-quantum[qiskit]"
    )

import ast
import json
import re
from azure.quantum import Job
from azure.quantum.qiskit.results.resource_estimator import ResourceEstimatorResult, ResourceEstimatorBatchResult

import logging
logger = logging.getLogger(__name__)

AzureJobStatusMap = {
    "Succeeded": JobStatus.DONE,
    "Waiting": JobStatus.QUEUED,
    "Executing": JobStatus.RUNNING,
    "Failed": JobStatus.ERROR,
    "Cancelled": JobStatus.CANCELLED,
    "Finishing": JobStatus.RUNNING
}

# Constants for output data format:
MICROSOFT_OUTPUT_DATA_FORMAT = "microsoft.quantum-results.v1"
MICROSOFT_OUTPUT_DATA_FORMAT_V2 = "microsoft.quantum-results.v2"
IONQ_OUTPUT_DATA_FORMAT = "ionq.quantum-results.v1"
QUANTINUUM_OUTPUT_DATA_FORMAT = "honeywell.quantum-results.v1"
RESOURCE_ESTIMATOR_OUTPUT_DATA_FORMAT = "microsoft.resource-estimates.v1"

class AzureQuantumJob(JobV1):
    def __init__(
        self,
        backend,
        azure_job=None,
        **kwargs
    ) -> None:
        """
            A Job running on Azure Quantum
        """
        if azure_job is None:
            azure_job = Job.from_input_data(
                workspace=backend.provider().get_workspace(),
                session_id=backend.get_latest_session_id(),
                **kwargs
            )

        self._azure_job = azure_job
        self._workspace = backend.provider().get_workspace()

        super().__init__(backend, self._azure_job.id, **kwargs)

    def job_id(self):
        """ This job's id."""
        return self._azure_job.id

    def id(self):
        """ This job's id."""
        return self._azure_job.id

    def refresh(self):
        """ Refreshes the job metadata from the server."""
        return self._azure_job.refresh()

    def submit(self):
        """ Submits the job for execution. """
        self._azure_job.submit()
        return

    def result(self, timeout=None, sampler_seed=None):
        """Return the results of the job."""
        self._azure_job.wait_until_completed(timeout_secs=timeout)

        success = self._azure_job.details.status == "Succeeded"
        results = self._format_results(sampler_seed=sampler_seed)

        result_dict = {
            "results" : results if isinstance(results, list) else [results],
            "job_id" : self._azure_job.details.id,
            "backend_name" : self._backend.name(),
            "backend_version" : self._backend.version,
            "qobj_id" : self._azure_job.details.name,
            "success" : success,
            "error_data" : None if self._azure_job.details.error_data is None else self._azure_job.details.error_data.as_dict()
        }

        result_type = Result
        if self._azure_job.details.output_data_format == RESOURCE_ESTIMATOR_OUTPUT_DATA_FORMAT:
            is_simple_job = results['success'] and not isinstance(results['data'], list)
            result_type = ResourceEstimatorResult if is_simple_job else ResourceEstimatorBatchResult

        return result_type.from_dict(result_dict)

    def cancel(self):
        """Attempt to cancel the job."""
        self._workspace.cancel_job(self._azure_job)

    def status(self):
        """Return the status of the job, among the values of ``JobStatus``."""
        self._azure_job.refresh()
        status = AzureJobStatusMap[self._azure_job.details.status]
        return status

    def queue_position(self):
        """Return the position of the job in the queue. Currently not supported."""
        return None

    def _shots_count(self):
        # Some providers use 'count', some other 'shots', give preference to 'count':
        input_params = self._azure_job.details.input_params
        options = self.backend().options
        shots = \
            input_params["count"] if "count" in input_params else \
            input_params["shots"] if "shots" in input_params else \
            options.get("count") if "count" in vars(options) else \
            options.get("shots")

        return shots

    def _format_results(self, sampler_seed=None) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """ Populates the results datastructures in a format that is compatible with qiskit libraries. """

        if (self._azure_job.details.output_data_format == MICROSOFT_OUTPUT_DATA_FORMAT_V2):
            return self._format_microsoft_v2_results()

        success = self._azure_job.details.status == "Succeeded"

        job_result = {
            "data": {},
            "success": success,
            "header": {},
        }

        if success:
            if (self._azure_job.details.output_data_format == MICROSOFT_OUTPUT_DATA_FORMAT):
                job_result["data"] = self._format_microsoft_results(sampler_seed=sampler_seed)
                
            elif (self._azure_job.details.output_data_format == IONQ_OUTPUT_DATA_FORMAT):
                job_result["data"] = self._format_ionq_results(sampler_seed=sampler_seed)

            elif (self._azure_job.details.output_data_format == QUANTINUUM_OUTPUT_DATA_FORMAT):
                job_result["data"] = self._format_quantinuum_results()

            else:
                job_result["data"] = self._format_unknown_results()

        job_result["header"] = self._azure_job.details.metadata
        if "metadata" in job_result["header"]:
            job_result["header"]["metadata"] = json.loads(job_result["header"]["metadata"])

        job_result["shots"] = self._shots_count()
        return job_result

    def _draw_random_sample(self, sampler_seed, probabilities, shots):
        _norm = sum(probabilities.values())
        if _norm != 1:
            if np.isclose(_norm, 1.0, rtol=1e-4):
                probabilities = {k: v/_norm for k, v in probabilities.items()}
            else:
                raise ValueError(f"Probabilities do not add up to 1: {probabilities}")
        if not sampler_seed:
            import hashlib
            id = self.job_id()
            sampler_seed = int(hashlib.sha256(id.encode('utf-8')).hexdigest(), 16) % (2**32 - 1)
        rand = np.random.RandomState(sampler_seed)
        rand_values = rand.choice(list(probabilities.keys()), shots, p=list(probabilities.values()))
        return dict(zip(*np.unique(rand_values, return_counts=True)))

    @staticmethod
    def _to_bitstring(k, num_qubits, meas_map):
        # flip bitstring to convert to little Endian
        bitstring = format(int(k), f"0{num_qubits}b")[::-1]
        # flip bitstring to convert back to big Endian
        return "".join([bitstring[n] for n in meas_map])[::-1]

    def _format_ionq_results(self, sampler_seed=None):
        """ Translate IonQ's histogram data into a format that can be consumed by qiskit libraries. """
        az_result = self._azure_job.get_results()
        shots = self._shots_count()

        if "num_qubits" not in self._azure_job.details.metadata:
            raise ValueError(f"Job with ID {self.id()} does not have the required metadata (num_qubits) to format IonQ results.")

        meas_map = json.loads(self._azure_job.details.metadata.get("meas_map")) if "meas_map" in self._azure_job.details.metadata else None
        num_qubits = self._azure_job.details.metadata.get("num_qubits")

        if not 'histogram' in az_result:
            raise "Histogram missing from IonQ Job results"

        counts = defaultdict(int)
        probabilities = defaultdict(int)
        for key, value in az_result['histogram'].items():
            bitstring = self._to_bitstring(key, num_qubits, meas_map) if meas_map else key
            probabilities[bitstring] += value

        if self.backend().configuration().simulator:
            counts = self._draw_random_sample(sampler_seed, probabilities, shots)
        else:
            counts = {bitstring: np.round(shots * value) for bitstring, value in probabilities.items()}

        return {"counts": counts, "probabilities": probabilities}

    @staticmethod
    def _qir_to_qiskit_bitstring(obj):
        """Convert the data structure from Azure into the "schema" used by Qiskit """
        if isinstance(obj, str) and not re.match(r"[\d\s]+$", obj):
            obj = ast.literal_eval(obj)

        if isinstance(obj, tuple):
            # the outermost implied container is a tuple, and each item is
            # associated with a classical register. Azure and Qiskit order the
            # registers in opposite directions, so reverse here to match.
            return " ".join([AzureQuantumJob._qir_to_qiskit_bitstring(term) for term in reversed(obj)])
        elif isinstance(obj, list):
            # a list is for an individual classical register
            return "".join([str(bit) for bit in obj])
        else:
            return str(obj)

    def _format_microsoft_results(self, sampler_seed=None):
        """ Translate Microsoft's job results histogram into a format that can be consumed by qiskit libraries. """
        az_result = self._azure_job.get_results()
        shots = self._shots_count()

        if not 'Histogram' in az_result:
            raise "Histogram missing from Job results"

        histogram = az_result['Histogram']
        counts = {}
        probabilities = {}
        # The Histogram serialization is odd entries are key and even entries values
        # Make sure we have even entries
        if (len(histogram) % 2) == 0:
            items = range(0, len(histogram), 2)
            for i in items:
                bitstring = AzureQuantumJob._qir_to_qiskit_bitstring(histogram[i])

                value = histogram[i + 1]
                probabilities[bitstring] = value
        else:
            raise "Invalid number of items in Job results' histogram."

        if self.backend().configuration().simulator:
            counts = self._draw_random_sample(sampler_seed, probabilities, shots)
        else:
            counts = {bitstring: np.round(shots * value) for bitstring, value in probabilities.items()}

        return {"counts": counts, "probabilities": probabilities}
    
    def _format_quantinuum_results(self):
        """ Translate Quantinuum's histogram data into a format that can be consumed by qiskit libraries. """
        az_result = self._azure_job.get_results()
        all_bitstrings = [
            bitstrings for classical_register, bitstrings 
            in az_result.items() if classical_register != "access_token"
        ]
        counts = {}
        combined_bitstrings = ["".join(bitstrings) for bitstrings in zip(*all_bitstrings)]
        shots = len(combined_bitstrings)

        for bitstring in set(combined_bitstrings):
            counts[bitstring] = combined_bitstrings.count(bitstring)

        histogram = {bitstring: count/shots for bitstring, count in counts.items()}

        return {"counts": counts, "probabilities": histogram}

    def _format_unknown_results(self):
        """ This method is called to format Job results data when the job output is in an unknown format."""
        az_result = self._azure_job.get_results()
        return az_result

    def _translate_microsoft_v2_results(self):
        """ Translate Microsoft's batching job results histograms into a format that can be consumed by qiskit libraries. """
        az_result = self._azure_job.get_results()

        if not "DataFormat" in az_result:
            raise ValueError("DataFormat missing from Job results")

        if not "Results" in az_result:
            raise ValueError("Results missing from Job results")

        histograms = []
        results = az_result["Results"]
        for circuit_results in results:
            counts = {}
            probabilities = {}

            if not "TotalCount" in circuit_results:
                raise ValueError("TotalCount missing from Job results")

            total_count = circuit_results["TotalCount"]

            if total_count <= 0:
                raise ValueError("TotalCount must be a positive non-zero integer")

            if not "Histogram" in circuit_results:
                raise ValueError("Histogram missing from Job results")
        
            histogram = circuit_results["Histogram"]
            for result in histogram:
                if not "Display" in result:
                    raise ValueError("Dispaly missing from histogram result")

                if not "Count" in result:
                    raise ValueError("Count missing from histogram result")

                bitstring = AzureQuantumJob._qir_to_qiskit_bitstring(result["Display"])
                count = result["Count"]
                probability = count / total_count
                counts[bitstring] = count
                probabilities[bitstring] = probability
            histograms.append((total_count, {"counts": counts, "probabilities": probabilities}))
        return histograms

    def _get_entry_point_names(self):
        input_params = self._azure_job.details.input_params
        # All V2 output is a list of entry points
        entry_points = input_params["items"]
        entry_point_names = []
        for entry_point in entry_points:
            if not "entryPoint" in entry_point:
                raise ValueError("Entry point input_param is missing an 'entryPoint' field")
            entry_point_names.append(entry_point["entryPoint"])
        return entry_point_names if len(entry_point_names) > 0 else ["main"]

    def _format_microsoft_v2_results(self) -> List[Dict[str, Any]]:
        success = self._azure_job.details.status == "Succeeded"

        if not success:
            return [{
                "data": {},
                "success": False,
                "header": {},
                "shots": 0,
            }]
        
        entry_point_names = self._get_entry_point_names()

        results = self._translate_microsoft_v2_results()
        
        if len(results) != len(entry_point_names):
            raise ValueError("The number of experiment results does not match the number of experiment names")
        
        status = self.status()

        return [{
            "data": result,
            "success": success,
            "shots": total_count,
            "name": name,
            "status": status
        } for name, (total_count, result) in zip(entry_point_names, results)]



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

import logging
from .version import __version__

from .job.job import *
from .job.session import *
from .workspace import *

from ._client.models._enums import JobStatus, SessionStatus, SessionJobFailurePolicy, ItemType

logger = logging.getLogger(__name__)
logger.info(f"version: {__version__}")



# coding=utf-8
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import warnings
warnings.warn("The azure.quantum.optimization.oneqbit namespace will be deprecated. \
Please use azure.quantum.target.oneqbit instead.")

from .solvers import (
    TabuSearch,
    PticmSolver,
    PathRelinkingSolver,
)



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

"""Azure Quantum Qiskit Provider"""

from .provider import AzureQuantumProvider
from azure.quantum import __version__

__all__ = [
    "AzureQuantumProvider",
    "__version__"
]



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from typing import Dict, List, TYPE_CHECKING, Union
from azure.quantum.aio.target import Target as AsyncTarget
from azure.quantum.aio.target import *
from azure.quantum.target.target_factory import TargetFactory as SyncTargetFactory

if TYPE_CHECKING:
    from azure.quantum.aio import Workspace

# Target ID keyword for parameter-free solvers
PARAMETER_FREE = "parameterfree"


class TargetFactory(SyncTargetFactory):
    """Factory class for generating a Target based on a
    provider and target name
    """
    __instance = None
    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls, *args, **kwargs)
        return cls.__instance

    @staticmethod
    def _get_all_target_cls() -> Dict[str, Target]:
        """Get all target classes by target name"""
        return {
            name: _t for t in AsyncTarget.__subclasses__()
            for _t in [t] + t.__subclasses__()
            for name in _t.target_names
        }

    async def get_targets(
        self,
        workspace: "Workspace",
        name: str,
        provider_id: str,
        **kwargs
    ) -> Union[Target, List[Target]]:
        """Create targets that are available to this workspace
        filtered by name and provider ID.

        :param name: Target name
        :type name: str
        :param workspace: Workspace
        :type workspace: Workspace
        :param provider_id: Provider name
        :type provider_id: str
        :return: One or more Target objects
        :rtype: Union[Target, List[Target]]
        """
        target_statuses = await workspace._get_target_status(name, provider_id, **kwargs)

        if len(target_statuses) == 1:
            _, status = target_statuses[0]
            return self.from_target_status(workspace, status=status)

        else:
            # Don't return redundant parameter-free targets
            return [
                self.from_target_status(_provider_id, status, **kwargs)
                for _provider_id, status in target_statuses
                if PARAMETER_FREE not in status.id
                and (
                    _provider_id.lower() in self._default_targets
                    or status.id in self._all_targets
                )
            ]




# pylint: disable=too-many-lines
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
from typing import Any, AsyncIterable, Callable, Dict, IO, List, Optional, TypeVar, Union, overload
import urllib.parse

from azure.core.async_paging import AsyncItemPaged, AsyncList
from azure.core.exceptions import (
    ClientAuthenticationError,
    HttpResponseError,
    ResourceExistsError,
    ResourceNotFoundError,
    ResourceNotModifiedError,
    map_error,
)
from azure.core.pipeline import PipelineResponse
from azure.core.pipeline.transport import AsyncHttpResponse
from azure.core.rest import HttpRequest
from azure.core.tracing.decorator import distributed_trace
from azure.core.tracing.decorator_async import distributed_trace_async
from azure.core.utils import case_insensitive_dict

from ... import models as _models
from ...operations._operations import (
    build_jobs_cancel_request,
    build_jobs_create_request,
    build_jobs_get_request,
    build_jobs_list_request,
    build_jobs_patch_request,
    build_providers_get_status_request,
    build_quotas_list_request,
    build_sessions_close_request,
    build_sessions_get_request,
    build_sessions_jobs_list_request,
    build_sessions_list_request,
    build_sessions_open_request,
    build_storage_sas_uri_request,
    build_top_level_items_list_request,
)

T = TypeVar("T")
ClsType = Optional[Callable[[PipelineResponse[HttpRequest, AsyncHttpResponse], T, Dict[str, Any]], Any]]


class JobsOperations:
    """
    .. warning::
        **DO NOT** instantiate this class directly.

        Instead, you should access the following operations through
        :class:`~azure.quantum._client.aio.QuantumClient`'s
        :attr:`jobs` attribute.
    """

    models = _models

    def __init__(self, *args, **kwargs) -> None:
        input_args = list(args)
        self._client = input_args.pop(0) if input_args else kwargs.pop("client")
        self._config = input_args.pop(0) if input_args else kwargs.pop("config")
        self._serialize = input_args.pop(0) if input_args else kwargs.pop("serializer")
        self._deserialize = input_args.pop(0) if input_args else kwargs.pop("deserializer")

    @distributed_trace
    def list(self, **kwargs: Any) -> AsyncIterable["_models.JobDetails"]:
        """List jobs.

        :return: An iterator like instance of JobDetails
        :rtype: ~azure.core.async_paging.AsyncItemPaged[~azure.quantum._client.models.JobDetails]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models._models.JobDetailsList] = kwargs.pop("cls", None)  # pylint: disable=protected-access

        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                request = build_jobs_list_request(
                    subscription_id=self._config.subscription_id,
                    resource_group_name=self._config.resource_group_name,
                    workspace_name=self._config.workspace_name,
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                request.url = self._client.format_url(request.url)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                request.url = self._client.format_url(request.url)

            return request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize(
                _models._models.JobDetailsList, pipeline_response  # pylint: disable=protected-access
            )
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
                raise HttpResponseError(response=response, model=error)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @distributed_trace_async
    async def get(self, job_id: str, **kwargs: Any) -> _models.JobDetails:
        """Get job by id.

        :param job_id: Id of the job. Required.
        :type job_id: str
        :return: JobDetails
        :rtype: ~azure.quantum._client.models.JobDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models.JobDetails] = kwargs.pop("cls", None)

        request = build_jobs_get_request(
            job_id=job_id,
            subscription_id=self._config.subscription_id,
            resource_group_name=self._config.resource_group_name,
            workspace_name=self._config.workspace_name,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        request.url = self._client.format_url(request.url)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
            raise HttpResponseError(response=response, model=error)

        deserialized = self._deserialize("JobDetails", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    @overload
    async def create(
        self, job_id: str, job: _models.JobDetails, *, content_type: str = "application/json", **kwargs: Any
    ) -> _models.JobDetails:
        """Create a job.

        :param job_id: Id of the job. Required.
        :type job_id: str
        :param job: The complete metadata of the job to submit. Required.
        :type job: ~azure.quantum._client.models.JobDetails
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JobDetails
        :rtype: ~azure.quantum._client.models.JobDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def create(
        self, job_id: str, job: IO, *, content_type: str = "application/json", **kwargs: Any
    ) -> _models.JobDetails:
        """Create a job.

        :param job_id: Id of the job. Required.
        :type job_id: str
        :param job: The complete metadata of the job to submit. Required.
        :type job: IO
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JobDetails
        :rtype: ~azure.quantum._client.models.JobDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def create(self, job_id: str, job: Union[_models.JobDetails, IO], **kwargs: Any) -> _models.JobDetails:
        """Create a job.

        :param job_id: Id of the job. Required.
        :type job_id: str
        :param job: The complete metadata of the job to submit. Is either a JobDetails type or a IO
         type. Required.
        :type job: ~azure.quantum._client.models.JobDetails or IO
        :keyword content_type: Body Parameter content-type. Known values are: 'application/json'.
         Default value is None.
        :paramtype content_type: str
        :return: JobDetails
        :rtype: ~azure.quantum._client.models.JobDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[_models.JobDetails] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(job, (IO, bytes)):
            _content = job
        else:
            _json = self._serialize.body(job, "JobDetails")

        request = build_jobs_create_request(
            job_id=job_id,
            subscription_id=self._config.subscription_id,
            resource_group_name=self._config.resource_group_name,
            workspace_name=self._config.workspace_name,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        request.url = self._client.format_url(request.url)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200, 201]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
            raise HttpResponseError(response=response, model=error)

        if response.status_code == 200:
            deserialized = self._deserialize("JobDetails", pipeline_response)

        if response.status_code == 201:
            deserialized = self._deserialize("JobDetails", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})  # type: ignore

        return deserialized  # type: ignore

    @distributed_trace_async
    async def cancel(self, job_id: str, **kwargs: Any) -> None:  # pylint: disable=inconsistent-return-statements
        """Cancel a job.

        :param job_id: Id of the job. Required.
        :type job_id: str
        :return: None
        :rtype: None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[None] = kwargs.pop("cls", None)

        request = build_jobs_cancel_request(
            job_id=job_id,
            subscription_id=self._config.subscription_id,
            resource_group_name=self._config.resource_group_name,
            workspace_name=self._config.workspace_name,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        request.url = self._client.format_url(request.url)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
            raise HttpResponseError(response=response, model=error)

        if cls:
            return cls(pipeline_response, None, {})

    @overload
    async def patch(
        self,
        job_id: str,
        patch_job: List[_models.JsonPatchDocument],
        *,
        content_type: str = "application/json",
        **kwargs: Any
    ) -> Optional[_models.JobDetails]:
        """Patch a job.

        :param job_id: Id of the job. Required.
        :type job_id: str
        :param patch_job: The json patch document containing the patch operations. Required.
        :type patch_job: list[~azure.quantum._client.models.JsonPatchDocument]
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JobDetails or None
        :rtype: ~azure.quantum._client.models.JobDetails or None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def patch(
        self, job_id: str, patch_job: IO, *, content_type: str = "application/json", **kwargs: Any
    ) -> Optional[_models.JobDetails]:
        """Patch a job.

        :param job_id: Id of the job. Required.
        :type job_id: str
        :param patch_job: The json patch document containing the patch operations. Required.
        :type patch_job: IO
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: JobDetails or None
        :rtype: ~azure.quantum._client.models.JobDetails or None
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def patch(
        self, job_id: str, patch_job: Union[List[_models.JsonPatchDocument], IO], **kwargs: Any
    ) -> Optional[_models.JobDetails]:
        """Patch a job.

        :param job_id: Id of the job. Required.
        :type job_id: str
        :param patch_job: The json patch document containing the patch operations. Is either a
         [JsonPatchDocument] type or a IO type. Required.
        :type patch_job: list[~azure.quantum._client.models.JsonPatchDocument] or IO
        :keyword content_type: Body Parameter content-type. Known values are: 'application/json'.
         Default value is None.
        :paramtype content_type: str
        :return: JobDetails or None
        :rtype: ~azure.quantum._client.models.JobDetails or None
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[Optional[_models.JobDetails]] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(patch_job, (IO, bytes)):
            _content = patch_job
        else:
            _json = self._serialize.body(patch_job, "[JsonPatchDocument]")

        request = build_jobs_patch_request(
            job_id=job_id,
            subscription_id=self._config.subscription_id,
            resource_group_name=self._config.resource_group_name,
            workspace_name=self._config.workspace_name,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        request.url = self._client.format_url(request.url)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200, 204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
            raise HttpResponseError(response=response, model=error)

        deserialized = None
        if response.status_code == 200:
            deserialized = self._deserialize("JobDetails", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized


class ProvidersOperations:
    """
    .. warning::
        **DO NOT** instantiate this class directly.

        Instead, you should access the following operations through
        :class:`~azure.quantum._client.aio.QuantumClient`'s
        :attr:`providers` attribute.
    """

    models = _models

    def __init__(self, *args, **kwargs) -> None:
        input_args = list(args)
        self._client = input_args.pop(0) if input_args else kwargs.pop("client")
        self._config = input_args.pop(0) if input_args else kwargs.pop("config")
        self._serialize = input_args.pop(0) if input_args else kwargs.pop("serializer")
        self._deserialize = input_args.pop(0) if input_args else kwargs.pop("deserializer")

    @distributed_trace
    def get_status(self, **kwargs: Any) -> AsyncIterable["_models.ProviderStatus"]:
        """Get provider status.

        :return: An iterator like instance of ProviderStatus
        :rtype: ~azure.core.async_paging.AsyncItemPaged[~azure.quantum._client.models.ProviderStatus]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models._models.ProviderStatusList] = kwargs.pop("cls", None)  # pylint: disable=protected-access

        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                request = build_providers_get_status_request(
                    subscription_id=self._config.subscription_id,
                    resource_group_name=self._config.resource_group_name,
                    workspace_name=self._config.workspace_name,
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                request.url = self._client.format_url(request.url)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                request.url = self._client.format_url(request.url)

            return request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize(
                _models._models.ProviderStatusList, pipeline_response  # pylint: disable=protected-access
            )
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
                raise HttpResponseError(response=response, model=error)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)


class StorageOperations:
    """
    .. warning::
        **DO NOT** instantiate this class directly.

        Instead, you should access the following operations through
        :class:`~azure.quantum._client.aio.QuantumClient`'s
        :attr:`storage` attribute.
    """

    models = _models

    def __init__(self, *args, **kwargs) -> None:
        input_args = list(args)
        self._client = input_args.pop(0) if input_args else kwargs.pop("client")
        self._config = input_args.pop(0) if input_args else kwargs.pop("config")
        self._serialize = input_args.pop(0) if input_args else kwargs.pop("serializer")
        self._deserialize = input_args.pop(0) if input_args else kwargs.pop("deserializer")

    @overload
    async def sas_uri(
        self, blob_details: _models.BlobDetails, *, content_type: str = "application/json", **kwargs: Any
    ) -> _models.SasUriResponse:
        """Gets a URL with SAS token for a container/blob in the storage account associated with the
        workspace. The SAS URL can be used to upload job input and/or download job output.

        :param blob_details: The details (name and container) of the blob to store or download data.
         Required.
        :type blob_details: ~azure.quantum._client.models.BlobDetails
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: SasUriResponse
        :rtype: ~azure.quantum._client.models.SasUriResponse
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def sas_uri(
        self, blob_details: IO, *, content_type: str = "application/json", **kwargs: Any
    ) -> _models.SasUriResponse:
        """Gets a URL with SAS token for a container/blob in the storage account associated with the
        workspace. The SAS URL can be used to upload job input and/or download job output.

        :param blob_details: The details (name and container) of the blob to store or download data.
         Required.
        :type blob_details: IO
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: SasUriResponse
        :rtype: ~azure.quantum._client.models.SasUriResponse
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def sas_uri(self, blob_details: Union[_models.BlobDetails, IO], **kwargs: Any) -> _models.SasUriResponse:
        """Gets a URL with SAS token for a container/blob in the storage account associated with the
        workspace. The SAS URL can be used to upload job input and/or download job output.

        :param blob_details: The details (name and container) of the blob to store or download data. Is
         either a BlobDetails type or a IO type. Required.
        :type blob_details: ~azure.quantum._client.models.BlobDetails or IO
        :keyword content_type: Body Parameter content-type. Known values are: 'application/json'.
         Default value is None.
        :paramtype content_type: str
        :return: SasUriResponse
        :rtype: ~azure.quantum._client.models.SasUriResponse
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[_models.SasUriResponse] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(blob_details, (IO, bytes)):
            _content = blob_details
        else:
            _json = self._serialize.body(blob_details, "BlobDetails")

        request = build_storage_sas_uri_request(
            subscription_id=self._config.subscription_id,
            resource_group_name=self._config.resource_group_name,
            workspace_name=self._config.workspace_name,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        request.url = self._client.format_url(request.url)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
            raise HttpResponseError(response=response, model=error)

        deserialized = self._deserialize("SasUriResponse", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized


class QuotasOperations:
    """
    .. warning::
        **DO NOT** instantiate this class directly.

        Instead, you should access the following operations through
        :class:`~azure.quantum._client.aio.QuantumClient`'s
        :attr:`quotas` attribute.
    """

    models = _models

    def __init__(self, *args, **kwargs) -> None:
        input_args = list(args)
        self._client = input_args.pop(0) if input_args else kwargs.pop("client")
        self._config = input_args.pop(0) if input_args else kwargs.pop("config")
        self._serialize = input_args.pop(0) if input_args else kwargs.pop("serializer")
        self._deserialize = input_args.pop(0) if input_args else kwargs.pop("deserializer")

    @distributed_trace
    def list(self, **kwargs: Any) -> AsyncIterable["_models.Quota"]:
        """List quotas for the given workspace.

        :return: An iterator like instance of Quota
        :rtype: ~azure.core.async_paging.AsyncItemPaged[~azure.quantum._client.models.Quota]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models._models.QuotaList] = kwargs.pop("cls", None)  # pylint: disable=protected-access

        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                request = build_quotas_list_request(
                    subscription_id=self._config.subscription_id,
                    resource_group_name=self._config.resource_group_name,
                    workspace_name=self._config.workspace_name,
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                request.url = self._client.format_url(request.url)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                request.url = self._client.format_url(request.url)

            return request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize(
                _models._models.QuotaList, pipeline_response  # pylint: disable=protected-access
            )
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
                raise HttpResponseError(response=response, model=error)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)


class SessionsOperations:
    """
    .. warning::
        **DO NOT** instantiate this class directly.

        Instead, you should access the following operations through
        :class:`~azure.quantum._client.aio.QuantumClient`'s
        :attr:`sessions` attribute.
    """

    models = _models

    def __init__(self, *args, **kwargs) -> None:
        input_args = list(args)
        self._client = input_args.pop(0) if input_args else kwargs.pop("client")
        self._config = input_args.pop(0) if input_args else kwargs.pop("config")
        self._serialize = input_args.pop(0) if input_args else kwargs.pop("serializer")
        self._deserialize = input_args.pop(0) if input_args else kwargs.pop("deserializer")

    @distributed_trace
    def list(self, **kwargs: Any) -> AsyncIterable["_models.SessionDetails"]:
        """List sessions.

        :return: An iterator like instance of SessionDetails
        :rtype: ~azure.core.async_paging.AsyncItemPaged[~azure.quantum._client.models.SessionDetails]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models._models.SessionDetailsList] = kwargs.pop("cls", None)  # pylint: disable=protected-access

        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                request = build_sessions_list_request(
                    subscription_id=self._config.subscription_id,
                    resource_group_name=self._config.resource_group_name,
                    workspace_name=self._config.workspace_name,
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                request.url = self._client.format_url(request.url)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                request.url = self._client.format_url(request.url)

            return request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize(
                _models._models.SessionDetailsList, pipeline_response  # pylint: disable=protected-access
            )
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
                raise HttpResponseError(response=response, model=error)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)

    @distributed_trace_async
    async def get(self, session_id: str, **kwargs: Any) -> _models.SessionDetails:
        """Get session by id.

        :param session_id: Id of the session. Required.
        :type session_id: str
        :return: SessionDetails
        :rtype: ~azure.quantum._client.models.SessionDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models.SessionDetails] = kwargs.pop("cls", None)

        request = build_sessions_get_request(
            session_id=session_id,
            subscription_id=self._config.subscription_id,
            resource_group_name=self._config.resource_group_name,
            workspace_name=self._config.workspace_name,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        request.url = self._client.format_url(request.url)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
            raise HttpResponseError(response=response, model=error)

        deserialized = self._deserialize("SessionDetails", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    @overload
    async def open(
        self, session_id: str, session: _models.SessionDetails, *, content_type: str = "application/json", **kwargs: Any
    ) -> _models.SessionDetails:
        """Open a session.

        :param session_id: Id of the session. Required.
        :type session_id: str
        :param session: The complete metadata of the session to be opened. Required.
        :type session: ~azure.quantum._client.models.SessionDetails
        :keyword content_type: Body Parameter content-type. Content type parameter for JSON body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: SessionDetails
        :rtype: ~azure.quantum._client.models.SessionDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @overload
    async def open(
        self, session_id: str, session: IO, *, content_type: str = "application/json", **kwargs: Any
    ) -> _models.SessionDetails:
        """Open a session.

        :param session_id: Id of the session. Required.
        :type session_id: str
        :param session: The complete metadata of the session to be opened. Required.
        :type session: IO
        :keyword content_type: Body Parameter content-type. Content type parameter for binary body.
         Default value is "application/json".
        :paramtype content_type: str
        :return: SessionDetails
        :rtype: ~azure.quantum._client.models.SessionDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """

    @distributed_trace_async
    async def open(
        self, session_id: str, session: Union[_models.SessionDetails, IO], **kwargs: Any
    ) -> _models.SessionDetails:
        """Open a session.

        :param session_id: Id of the session. Required.
        :type session_id: str
        :param session: The complete metadata of the session to be opened. Is either a SessionDetails
         type or a IO type. Required.
        :type session: ~azure.quantum._client.models.SessionDetails or IO
        :keyword content_type: Body Parameter content-type. Known values are: 'application/json'.
         Default value is None.
        :paramtype content_type: str
        :return: SessionDetails
        :rtype: ~azure.quantum._client.models.SessionDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = case_insensitive_dict(kwargs.pop("headers", {}) or {})
        _params = kwargs.pop("params", {}) or {}

        content_type: Optional[str] = kwargs.pop("content_type", _headers.pop("Content-Type", None))
        cls: ClsType[_models.SessionDetails] = kwargs.pop("cls", None)

        content_type = content_type or "application/json"
        _json = None
        _content = None
        if isinstance(session, (IO, bytes)):
            _content = session
        else:
            _json = self._serialize.body(session, "SessionDetails")

        request = build_sessions_open_request(
            session_id=session_id,
            subscription_id=self._config.subscription_id,
            resource_group_name=self._config.resource_group_name,
            workspace_name=self._config.workspace_name,
            content_type=content_type,
            api_version=self._config.api_version,
            json=_json,
            content=_content,
            headers=_headers,
            params=_params,
        )
        request.url = self._client.format_url(request.url)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200, 201]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
            raise HttpResponseError(response=response, model=error)

        if response.status_code == 200:
            deserialized = self._deserialize("SessionDetails", pipeline_response)

        if response.status_code == 201:
            deserialized = self._deserialize("SessionDetails", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})  # type: ignore

        return deserialized  # type: ignore

    @distributed_trace_async
    async def close(self, session_id: str, **kwargs: Any) -> _models.SessionDetails:
        """Close a session.

        :param session_id: Id of the session. Required.
        :type session_id: str
        :return: SessionDetails
        :rtype: ~azure.quantum._client.models.SessionDetails
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models.SessionDetails] = kwargs.pop("cls", None)

        request = build_sessions_close_request(
            session_id=session_id,
            subscription_id=self._config.subscription_id,
            resource_group_name=self._config.resource_group_name,
            workspace_name=self._config.workspace_name,
            api_version=self._config.api_version,
            headers=_headers,
            params=_params,
        )
        request.url = self._client.format_url(request.url)

        _stream = False
        pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
            request, stream=_stream, **kwargs
        )

        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
            raise HttpResponseError(response=response, model=error)

        deserialized = self._deserialize("SessionDetails", pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    @distributed_trace
    def jobs_list(self, session_id: str, **kwargs: Any) -> AsyncIterable["_models.JobDetails"]:
        """List jobs in a session.

        :param session_id: Id of the session. Required.
        :type session_id: str
        :return: An iterator like instance of JobDetails
        :rtype: ~azure.core.async_paging.AsyncItemPaged[~azure.quantum._client.models.JobDetails]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models._models.JobDetailsList] = kwargs.pop("cls", None)  # pylint: disable=protected-access

        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                request = build_sessions_jobs_list_request(
                    session_id=session_id,
                    subscription_id=self._config.subscription_id,
                    resource_group_name=self._config.resource_group_name,
                    workspace_name=self._config.workspace_name,
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                request.url = self._client.format_url(request.url)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                request.url = self._client.format_url(request.url)

            return request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize(
                _models._models.JobDetailsList, pipeline_response  # pylint: disable=protected-access
            )
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
                raise HttpResponseError(response=response, model=error)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)


class TopLevelItemsOperations:
    """
    .. warning::
        **DO NOT** instantiate this class directly.

        Instead, you should access the following operations through
        :class:`~azure.quantum._client.aio.QuantumClient`'s
        :attr:`top_level_items` attribute.
    """

    models = _models

    def __init__(self, *args, **kwargs) -> None:
        input_args = list(args)
        self._client = input_args.pop(0) if input_args else kwargs.pop("client")
        self._config = input_args.pop(0) if input_args else kwargs.pop("config")
        self._serialize = input_args.pop(0) if input_args else kwargs.pop("serializer")
        self._deserialize = input_args.pop(0) if input_args else kwargs.pop("deserializer")

    @distributed_trace
    def list(self, **kwargs: Any) -> AsyncIterable["_models.ItemDetails"]:
        """List top-level items.

        :return: An iterator like instance of ItemDetails
        :rtype: ~azure.core.async_paging.AsyncItemPaged[~azure.quantum._client.models.ItemDetails]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[_models._models.ItemDetailsList] = kwargs.pop("cls", None)  # pylint: disable=protected-access

        error_map = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                request = build_top_level_items_list_request(
                    subscription_id=self._config.subscription_id,
                    resource_group_name=self._config.resource_group_name,
                    workspace_name=self._config.workspace_name,
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                request.url = self._client.format_url(request.url)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version
                request = HttpRequest(
                    "GET", urllib.parse.urljoin(next_link, _parsed_next_link.path), params=_next_request_params
                )
                request.url = self._client.format_url(request.url)

            return request

        async def extract_data(pipeline_response):
            deserialized = self._deserialize(
                _models._models.ItemDetailsList, pipeline_response  # pylint: disable=protected-access
            )
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)  # type: ignore
            return deserialized.next_link or None, AsyncList(list_of_elem)

        async def get_next(next_link=None):
            request = prepare_request(next_link)

            _stream = False
            pipeline_response: PipelineResponse = await self._client._pipeline.run(  # pylint: disable=protected-access
                request, stream=_stream, **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.RestError, pipeline_response)
                raise HttpResponseError(response=response, model=error)

            return pipeline_response

        return AsyncItemPaged(get_next, extract_data)



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from azure.quantum.job.job import Job
from azure.quantum.job.workspace_item import WorkspaceItem
from azure.quantum.job.workspace_item_factory import WorkspaceItemFactory
from azure.quantum.job.session import Session, SessionHost
from azure.quantum._client.models import JobDetails



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging

from typing import Optional

from azure.quantum import Workspace
from azure.quantum.target.solvers import RangeSchedule, Solver
from azure.quantum.job.base_job import ContentType

logger = logging.getLogger(__name__)

class PopulationAnnealing(Solver):
    target_names = (
        "microsoft.populationannealing.cpu",
        "microsoft.populationannealing-parameterfree.cpu",
    )
    def __init__(
        self,
        workspace: Workspace,
        name: str = "microsoft.populationannealing.cpu",
        alpha: Optional[float] = None,
        seed: Optional[int] = None,
        population: Optional[int] = None,
        sweeps: Optional[int] = None,
        beta: Optional[RangeSchedule] = None,
        timeout: Optional[int] = None,
        content_type: Optional[ContentType] = ContentType.json,
        **kwargs
    ):
        """Constructor of the Population Annealing solver.

        Population Annealing Search solver for binary optimization problems
        with k-local interactions on an all-to-all graph topology with double
        precision support for the coupler weights.

        This solver is CPU only.

        :param alpha:
            Ratio to trigger a restart. Must be larger than 1.
        :param seed:
            Specifies the random number generator seed value.
        :population:
            Size of the population. Must be positive.
        :param sweeps:
            Number of Monte Carlo sweeps. Must be positive.
        :param beta:
            Evolution of the inverse annealing temperature.
            Must be an object of type RangeSchedule describing
            an increasing evolution (0 < initial < final).
        :param timeout:
            specifies maximum number of seconds to run the core solver
            loop. initialization time does not respect this value, so the
            solver may run longer than the value specified. Setting this value
            will trigger the parameter free population annealing solver.
        """

        if timeout is not None:
            name = "microsoft.populationannealing-parameterfree.cpu"

        super().__init__(
            workspace=workspace,
            provider_id="Microsoft",
            name=name,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2",
            content_type= content_type,
            **kwargs
        )

        self.check_set_float("alpha", alpha, lower_bound_exclusive=1.0)
        self.set_one_param("seed", seed)
        self.check_set_positive_int("population", population)
        self.check_set_positive_int("sweeps", sweeps)
        self.check_set_schedule(
                "beta", beta, evolution=self.ScheduleEvolution.INCREASING,
                lower_bound_exclusive=0)
        self.check_set_positive_int("timeout", timeout)

    def supports_grouped_terms(self):
        return True
    
    def supports_protobuf(self):
        return True